/* hdbapp_dump.c
** hdb dump: write out contents of hdb file
** wcm, 2010.05.30 - 2010.12.09
** ===
*/

/* libc: */
#include <stdlib.h>

/* unix: */
#include <unistd.h>
#include <errno.h>

/* lasagna: */
#include "buf.h"
#include "cstr.h"
#include "die.h"
#include "ioq.h"
#include "ioq_std.h"
#include "nextopt.h"
#include "nfmt.h"
#include "sysstr.h"

/* lasagna hdb reader: */
#include "hdb.h"

/* hdbapp: */
#include "hdbapp.h"


#ifndef HDBAPP_MODULE
/* configuration for standalone program: */
/* these globals setup in main(): */
const char    *progname = NULL;
const char    *usage_prog = NULL;
/* these globals remain NULL in standalone: */
const char    *progmode = NULL;
const char    *usage_mode = NULL;
/* entry point for standalone: */
#define  HDBAPP_MAIN  main
#else
/* entry point for module: */
#define  HDBAPP_MAIN  hdbapp_dump
#endif                          /* #ifndef HDBAPP_MODULE */


/* objects in local scope: */
static uchar_t *map = NULL;
static uint32_t mapsize = 0;
static char    nfmt[NFMT_SIZE];
#define  DUMPBUF  256
static uchar_t dumpbuf[DUMPBUF];

/* -y: cdb "legacy" output format */
static int     opt_y = 0;
/* -g: "getln" output format */
static int     opt_g = 0;
/* -d arg: delimiter for -g format */
static char   *arg_d = "\t";


/* functions in local scope: */
static void    do_put_map(struct hdb *H, uint32_t offset, size_t len);
static void    do_put_buf(struct hdb *H, uint32_t offset, size_t len);
static int     do_dump(struct hdb *H);


/* do_put_map()
**   direct copy from non-NULL map to ioq1
**   abort on failure
*/
static
void do_put_map(struct hdb *NOTUSED, uint32_t offset, size_t len)
{
   /* silence compiler warning unused parameter (nv): */
   (void) NOTUSED;

   if ((offset > mapsize) || ((mapsize - offset) < len)) {
      errno = EPROTO;
      fatal_syserr("map offset request exceeds map boundary");
   }
   ioq_put(ioq1, map + offset, len);

   return;
}

/* do_put_buf()
**   copy from hdb to ioq1 through intermediate dumpbuf
**   abort on failure
*/
static
void do_put_buf(struct hdb *H, uint32_t offset, size_t len)
{
   size_t         n = sizeof dumpbuf;

   while (len) {
      if (len < n)
         n = len;
      if (hdb_read(H, dumpbuf, n, offset) == -1) {
         fatal_syserr("failure during hdb_read()");
      }
      ioq_put(ioq1, dumpbuf, n);
      offset += n;
      len -= n;
   }

   return;
}


/* do_dump()
**   iterate and output each record in hdb 
*/
static
int do_dump(struct hdb *H)
{
   uint32_t       klen, dlen, offset;
   void           (*do_put) (struct hdb *, uint32_t, size_t);
   int            r;

   /* setup for either direct map or intermediate dumpbuf operation: */
   map = hdb_map(H);
   if (map) {
      mapsize = hdb_mapsize(H);
      do_put = &do_put_map;
   } else {
      do_put = &do_put_buf;
   }

   /* sequential iteration: */
   for (r = hdb_seqinit(H); r != 0; r = hdb_seqnext(H)) {
      if (r == -1) {
         fatal_syserr("failure during hdb_seq()");
      }

      klen = hdb_klen(H);
      dlen = hdb_dlen(H);

      if (!opt_g) {
         if (opt_y)
            ioq_puts(ioq1, "+");
         ioq_vputs(ioq1, nfmt_uint32(nfmt, klen), opt_y ? "," : ":");
         ioq_vputs(ioq1, nfmt_uint32(nfmt, dlen), opt_y ? ":" : "\t");
      }

      offset = hdb_kpos(H);
      do_put(H, offset, klen);
      if (!opt_g) {
         ioq_puts(ioq1, opt_y ? "->" : ":");
      } else {
         ioq_put(ioq1, (uchar_t *) arg_d, 1);
      }

      offset += klen;
      do_put(H, offset, dlen);
      ioq_puts(ioq1, "\n");
   }

   /* append final empty line for cdb legacy format: */
   if (opt_y)
      ioq_puts(ioq1, "\n");

   ioq_flush(ioq1);
   return 0;
}


int HDBAPP_MAIN(int argc, char *argv[])
{
   nextopt_t      nopt;
   char           opt;
   struct hdb     H;
   int            is_stdin = 0;
   int            e;

#ifndef HDBAPP_MODULE
/* standalone options/initialization: */
   usage_prog = "[-hV] [-d del | -g | -y] [hdb]";
   nextopt_init(&nopt, argc, argv, ":hVd:gy");
   progname = nextopt_progname(&nopt);
#else
/* module options/initialization: */
   usage_mode = "[-h] [-d del | -g | -y] [hdb]";
   nextopt_init(&nopt, argc, argv, ":hcd:gy");
#endif

   while ((opt = nextopt(&nopt))) {
      char           optc[2] = { nopt.opt_got, '\0' };
      switch (opt) {
      case 'h':
         usage();
         die(0);
         break;
      case 'V':
         version();
         die(0);
         break;
      case 'g':
         ++opt_g;
         break;
      case 'd':
         ++opt_g;
         arg_d = nopt.opt_arg;
         break;
      case 'y':
         ++opt_y;
         break;
      case ':':
         fatal_usage("missing argument for option -", optc);
         break;
      case '?':
         if (nopt.opt_got != '?') {
            fatal_usage("invalid option -", optc);
         }
         /* else fallthrough: */
      default:
         die_usage();
         break;
      }
   }

   argc -= nopt.arg_ndx;
   argv += nopt.arg_ndx;

   if (opt_g && opt_y) {
      fatal_usage("options -g and -y are mutually exclusive");
   }

   if ((argc > 0) && (argv[0] != NULL)) {
      /* hdb file argument: */
      if (hdb_open(&H, argv[0]) == -1) {
         fatal_syserr("failure initializing hdb from ", argv[0]);
      }
   } else {
      /* read from stdin: */
      if (hdb_init(&H, 0) == -1) {
         fatal_syserr("failure initializing hdb on stdin");
      }
      ++is_stdin;
   }

   /* note: do_dump() just aborts on failure */
   e = do_dump(&H);

   /* close/cleanup: */
   if (is_stdin) {
      hdb_clear(&H);
   } else {
      hdb_close(&H);
   }

   return e;
}


/* eof (hdbapp_dump.c) */
