/* hdbapp_dupes.c
** hdb dupes module
** wcm, 2010.11.02 - 2010.12.14
** ===
*/

/* libc: */
#include <stdlib.h>

/* unix: */
#include <unistd.h>
#include <errno.h>

/* lasagna: */
#include "buf.h"
#include "cstr.h"
#include "ioq.h"
#include "ioq_std.h"
#include "nextopt.h"
#include "nfmt.h"
#include "sysstr.h"
#include "uchar.h"
#include "upak.h"

/* lasagna hdb reader: */
#include "hdb.h"

/* hdbapp: */
#include "hdbapp.h"


#ifndef HDBAPP_MODULE
/* configuration for standalone program: */
/* these globals setup in main(): */
const char    *progname = NULL;
const char    *usage_prog = NULL;
/* these globals remain NULL in standalone: */
const char    *progmode = NULL;
const char    *usage_mode = NULL;
/* entry point for standalone: */
#define  HDBAPP_MAIN  main
#else
/* entry point for module: */
#define  HDBAPP_MAIN  hdbapp_dupes
#endif                          /* #ifndef HDBAPP_MODULE */


/*
** objects in local scope - options:
*/

/* -y: cdb "legacy" output format */
static int     opt_y = 0;
/* -g: "getline" output format */
static int     opt_g = 0;
/* -d arg: delimiter for -g format */
static char   *arg_d = "\t";
/* -q: "quick/quiet" mode */
static int     opt_q = 0;

/*
** other objects in local scope:
*/
static uchar_t *map = NULL;
static uint32_t mapsize = 0;
static char    nfmt[NFMT_SIZE];
#define  DUMPBUF  256
static uchar_t dumpbuf[DUMPBUF];

/* do_put() function pointer: */
void           (*do_put) (struct hdb * H, uint32_t offset, size_t len) = NULL;


/*
** functions in local scope:
*/
static void    do_put_map(struct hdb *H, uint32_t offset, size_t len);
static void    do_put_buf(struct hdb *H, uint32_t offset, size_t len);
static void    do_output(struct hdb *H);
static int     do_dupes(struct hdb *H);


/* do_put_map()
**   direct copy from non-NULL map to ioq1
**   abort on failure
*/
static
void do_put_map(struct hdb *NOTUSED, uint32_t offset, size_t len)
{
   /* silence compiler warning unused parameter (nv): */
   (void) NOTUSED;

   if ((offset > mapsize) || ((mapsize - offset) < len)) {
      errno = EPROTO;
      fatal_syserr("map offset request exceeds map boundary");
   }
   ioq_put(ioq1, map + offset, len);

   return;
}


/* do_put_buf()
**   copy from hdb to ioq1 through intermediate dumpbuf
**   abort on failure
*/
static
void do_put_buf(struct hdb *H, uint32_t offset, size_t len)
{
   size_t         n = sizeof dumpbuf;

   while (len) {
      if (len < n)
         n = len;
      if (hdb_read(H, dumpbuf, n, offset) == -1) {
         fatal_syserr("failure during hdb_read()");
      }
      ioq_put(ioq1, dumpbuf, n);
      offset += n;
      len -= n;
   }

   return;
}


/* do_output()
**   output a formatted record for current cursor in hdb H
**   abort on failure
*/
static
void do_output(struct hdb *H)
{
   uint32_t       klen = hdb_klen(H);
   uint32_t       dlen = hdb_dlen(H);

   if (!opt_g) {
      if (opt_y)
         ioq_puts(ioq1, "+");
      ioq_vputs(ioq1, nfmt_uint32(nfmt, klen), opt_y ? "," : ":");
      ioq_vputs(ioq1, nfmt_uint32(nfmt, dlen), opt_y ? ":" : "\t");
   }

   /* key: */
   do_put(H, hdb_kpos(H), klen);

   if (!opt_g) {
      ioq_puts(ioq1, opt_y ? "->" : ":");
   } else {
      ioq_put(ioq1, (uchar_t *) arg_d, 1);
   }

   /* data: */
   do_put(H, hdb_dpos(H), dlen);

   ioq_puts(ioq1, "\n");

   return;
}


/* do_dupes()
**   sequentially scan hdb for duplicate keys
**   output records for any duplicate keys found
**   abort on failure
**   return:
**     0 : successful completion
**     1 : in -q "quick/quiet" mode, dupes found (no output)
*/
static
int do_dupes(struct hdb *H)
{
   struct hdb     CC;
   uint32_t       rpos, kpos, klen;
   dynbuf         K = dynbuf_INIT();
   int            shown_first;
   uint32_t       dupekeys = 0, duperecs = 0;
   int            e, r;

   /* setup for either direct map or intermediate dumpbuf output operation: */
   map = hdb_map(H);
   if (map) {
      mapsize = hdb_mapsize(H);
      do_put = do_put_map;
   } else {
      do_put = do_put_buf;
   }

   /* make copy of hdb object for sequential iterator: */
   hdb_cc(&CC, H);

   /* scan records: */
   for (r = hdb_seqinit(&CC); r != 0; r = hdb_seqnext(&CC)) {
      if (r == -1) {
         fatal_syserr("failure during hdb_seq()");
      }

      rpos = hdb_rpos(&CC);
      klen = hdb_klen(&CC);
      kpos = hdb_kpos(&CC);

      dynbuf_CLEAR(&K);
      e = hdb_dynread(H, &K, klen, kpos);
      if (e == -1) {
         fatal_syserr("failure during hdb_dynread()");
      }

      /* hdb_find() must succeed here: */
      switch (hdb_find(H, dynbuf_BUF(&K), klen)) {
      case -1:
         fatal_syserr("failure during hdb_find()");
         break;
      case 0:
         errno = EPROTO;
         fatal_syserr("hdb file integrity error");
         break;
      }

      if (rpos != hdb_rpos(H)) {
         /*
          ** initial hdb_find() will return first record in any dupe sequence
          ** if this is not the first record in a dupe sequence, the dupe
          ** sequence has already been considered, and we can iterate to
          ** next record
          */
         continue;
      }

      shown_first = 0;
      /* look for dupes on current key: */
      for (;;) {
         e = hdb_findnext(H);
         if (e == -1)
            fatal_syserr("failure during hdb_findnext()");
         if (e == 0)
            break;

         /* else dupe found: */
         if (opt_q) {
            /* quick/quiet: */
            dynbuf_freebuf(&K);
            return 1;
         }

         if (!shown_first) {
            /* display first record in dupe sequence: */
            do_output(&CC);
            ++shown_first;
            ++dupekeys;
         }
         /* display this record: */
         do_output(H);
         ++duperecs;
      }
   }

   if (opt_q) {
      /* quick/quiet: no dupes found */
      return 0;
   }

   /* append final empty line for cdb legacy format: */
   if (opt_y)
      ioq_puts(ioq1, "\n");


   /* all done with stdout: */
   ioq_flush(ioq1);
   dynbuf_freebuf(&K);

   /* summary report to stderr: */
   if (progmode) {
      ioq_vputs(ioq2, progname, " ", progmode, ": ");
   } else {
      ioq_vputs(ioq2, progname, ": ");
   }
   if (dupekeys) {
      ioq_vputs(ioq2, nfmt_uint32(nfmt, dupekeys), " duplicate key",
                (dupekeys != 1) ? "s" : "");
      ioq_vputs(ioq2, "across ", nfmt_uint32(nfmt, duperecs), "records\n");
   } else {
      ioq_puts(ioq2, "no duplicate keys found\n");
   }
   ioq_flush(ioq2);

   return 0;
}


int HDBAPP_MAIN(int argc, char *argv[])
{
   nextopt_t      nopt;
   char           opt;
   struct hdb     H;
   int            is_stdin = 0;
   int            e;

#ifndef HDBAPP_MODULE
/* standalone options/initialization: */
   usage_prog = "[-hV] [-d del | -g | -y] [-q] [hdb]";
   nextopt_init(&nopt, argc, argv, ":hVd:gyq");
   progname = nextopt_progname(&nopt);
#else
/* module options/initialization: */
   usage_mode = "[-h] [-d del | -g | -y] [-q] [hdb]";
   nextopt_init(&nopt, argc, argv, ":hd:gyq");
#endif

   while ((opt = nextopt(&nopt))) {
      char           optc[2] = { nopt.opt_got, '\0' };
      switch (opt) {
      case 'h':
         usage();
         die(0);
         break;
      case 'V':
         version();
         die(0);
         break;
      case 'd':
         ++opt_g;
         arg_d = nopt.opt_arg;
         break;
      case 'g':
         ++opt_g;
         break;
      case 'y':
         ++opt_y;
         break;
      case 'q':
         ++opt_q;
         break;
      case ':':
         fatal_usage("missing argument for option -", optc);
         break;
      case '?':
         if (nopt.opt_got != '?') {
            fatal_usage("invalid option -", optc);
         }
         /* else fallthrough: */
      default:
         die_usage();
         break;
      }
   }

   argc -= nopt.arg_ndx;
   argv += nopt.arg_ndx;

   if (opt_g && opt_y) {
      fatal_usage("options -g and -y are mutually exclusive");
   }

   if ((argc > 0) && (argv[0] != NULL)) {
      /* hdb file argument: */
      if (hdb_open(&H, argv[0]) == -1) {
         fatal_syserr("failure initializing hdb from ", argv[0]);
      }
   } else {
      /* read from stdin: */
      if (hdb_init(&H, 0) == -1) {
         fatal_syserr("failure initializing hdb on stdin");
      }
      ++is_stdin;
   }

   /* note: do_dupes() just aborts on failure */
   e = do_dupes(&H);

   /* cleanup/close: */
   if (is_stdin) {
      hdb_clear(&H);
   } else {
      hdb_close(&H);
   }

   return e;
}


/* eof (hdbapp_dupes.c) */
