/* hdbapp_get.c
** hdb get: find key in hdb
** wcm, 2010.05.30 - 2010.12.14
** ===
*/

/* libc: */
#include <stdlib.h>

/* unix: */
#include <unistd.h>
#include <errno.h>

/* lasagna: */
#include "cstr.h"
#include "dynbuf.h"
#include "ioq.h"
#include "ioq_std.h"
#include "nextopt.h"
#include "nuscan.h"
#include "sysstr.h"
#include "uchar.h"

/* lasagna hdb reader: */
#include "hdb.h"

/* hdbapp: */
#include "hdbapp.h"


#ifndef HDBAPP_MODULE
/* configuration for standalone program: */
/* these globals setup in main(): */
const char    *progname = NULL;
const char    *usage_prog = NULL;
/* these globals remain NULL in standalone: */
const char    *progmode = NULL;
const char    *usage_mode = NULL;
/* entry point for standalone: */
#define  HDBAPP_MAIN  main
#else
/* entry point for module: */
#define  HDBAPP_MAIN  hdbapp_get
#endif                          /* #ifndef HDBAPP_MODULE */


/* objects in local scope: */
/* -a: "all" */
static int     opt_a = 0;
/* -n: "omit trailing newline" */
static int     opt_n = 0;
/* -j arg: "jump" */
static int     opt_j = 0;
static uint32_t arg_j = 0;


/* functions in local scope: */
static int     do_get(struct hdb *H, const char *key);


/* do_get()
**   lookup key in hdb
**   return:
**     0: found
**     1: not found
**
**   abort on failure
*/
static
int do_get(struct hdb *H, const char *key)
{
   uint32_t       klen;
   dynbuf         D = dynbuf_INIT();
   int            found;
   uint32_t       jump = opt_j ? arg_j : 0;
   uint32_t       n;

   klen = cstr_len(key);

   for (found = hdb_find(H, (uchar_t *) key, klen), n = 0;
        found > 0; found = hdb_findnext(H), ++n) {
      /* show conditions: */
      if (opt_a || (opt_j && (n == jump)) || (!opt_a && !opt_j && (n == 0))) {
         dynbuf_CLEAR(&D);
         if (hdb_dynget(H, &D) == -1) {
            fatal_syserr("failure hdb_dynget()");
         }
         ioq_put(ioq1, dynbuf_BUF(&D), dynbuf_LEN(&D));
         if (!opt_n) {
            ioq_puts(ioq1, "\n");
         }
      }

      /* stop conditions: */
      if ((opt_j && (n == jump)) || (!opt_a && !opt_j && (n == 0))) {
         break;
      }
   }
   ioq_flush(ioq1);

   if (found == -1) {
      fatal_syserr("failure during hdb_find()");
   }

   if (found || (opt_a && n)) {
      return 0;
   }

   /* else key not found: */
   return 1;
}


int HDBAPP_MAIN(int argc, char *argv[])
{
   nextopt_t      nopt;
   char           opt;
   struct hdb     H;
   char          *key;
   const char    *z;
   int            is_stdin = 0;
   int            e;

#ifndef HDBAPP_MODULE
/* standalone options/initialization: */
   usage_prog = "[-hV] [-a | -j num] [-n] key [hdb]";
   nextopt_init(&nopt, argc, argv, ":hVaj:n");
   progname = nextopt_progname(&nopt);
#else
/* module options/initialization: */
   usage_mode = "[-h] [-a | -j num] [-n] key [hdb]";
   nextopt_init(&nopt, argc, argv, ":haj:n");
#endif

   while ((opt = nextopt(&nopt))) {
      char           optc[2] = { nopt.opt_got, '\0' };
      switch (opt) {
      case 'h':
         usage();
         die(0);
         break;
      case 'V':
         version();
         die(0);
         break;
      case 'a':
         ++opt_a;
         break;
      case 'j':
         ++opt_j;
         z = nuscan_uint32(&arg_j, nopt.opt_arg);
         if (*z != '\0') {
            fatal_usage("non-numeric argument found for option -", optc, ": ",
                        nopt.opt_arg);
         }
         break;
      case 'n':
         ++opt_n;
         break;
      case ':':
         fatal_usage("missing argument for option -", optc);
         break;
      case '?':
         if (nopt.opt_got != '?') {
            fatal_usage("invalid option -", optc);
         }
         /* else fallthrough: */
      default:
         die_usage();
         break;
      }
   }

   argc -= nopt.arg_ndx;
   argv += nopt.arg_ndx;

   if (opt_a && opt_j) {
      fatal_usage("options -a and -j are mutually exclusive");
   }

   if ((argc < 1) || ((key = argv[0]) == NULL)) {
      fatal_usage("missing search key argument");
   }
   --argc;
   ++argv;

   if ((argc > 0) && (argv[0] != NULL)) {
      /* hdb file argument: */
      if (hdb_open(&H, argv[0]) == -1) {
         fatal_syserr("failure initializing hdb from ", argv[0]);
      }
   } else {
      /* read from stdin: */
      if (hdb_init(&H, 0) == -1) {
         fatal_syserr("failure initializing hdb on stdin");
      }
      ++is_stdin;
   }


   /* do_get() returns 0 if found, 1 if not found; aborts on error */
   e = do_get(&H, key);

   /* close/cleanup: */
   if (is_stdin) {
      hdb_clear(&H);
   } else {
      hdb_close(&H);
   }

   return e;
}


/* eof (hdbapp_get.c) */
