/* hdbapp_grep.c
** hdb grep: regex match scan through hdb file
** wcm, 2010.11.01 - 2010.12.14
** ===
*/

/* libc: */
#include <stdlib.h>

/* unix: */
#include <unistd.h>
#include <errno.h>

/* regex: */
#include <regex.h>

/* lasagna: */
#include "cstr.h"
#include "die.h"
#include "ioq.h"
#include "ioq_std.h"
#include "nextopt.h"
#include "nfmt.h"
#include "sysstr.h"

/* lasagna hdb reader: */
#include "hdb.h"

/* hdbapp: */
#include "hdbapp.h"


#ifndef HDBAPP_MODULE
/* configuration for standalone program: */
/* these globals setup in main(): */
const char    *progname = NULL;
const char    *usage_prog = NULL;
/* these globals remain NULL in standalone: */
const char    *progmode = NULL;
const char    *usage_mode = NULL;
/* entry point for standalone: */
#define  HDBAPP_MAIN  main
#else
/* entry point for module: */
#define  HDBAPP_MAIN  hdbapp_grep
#endif                          /* #ifndef HDBAPP_MODULE */

/*
** objects in local scope - options:
*/

/* -i: case insensitive match */
static int     opt_i = 0;
/* -k: match on keys */
static int     opt_k = 0;
/* -!: invert match */
static int     opt_NOT = 0;

/* -y: cdb "legacy" output format */
static int     opt_y = 0;
/* -g: "getline" output format */
static int     opt_g = 0;
/* -d arg: delimiter for -g format */
static char   *arg_d = "\t";

/*
** objects in local scope - other :
*/

/* compiled regex: */
static regex_t rx;

/* stringified numerics: */
static char    nfmt[NFMT_SIZE];

/* functions in local scope: */
static int     do_grep(struct hdb *H);


/* do_grep()
**   iterate and grep each record
**   abort on fail
**
**   return:
**     0 : success, 1 or more matches found
**     1 : no match found
*/
static
int do_grep(struct hdb *H)
{
   uint32_t       klen, dlen;
   dynbuf         K = dynbuf_INIT();
   dynbuf         D = dynbuf_INIT();
   char          *key, *data;
   uint32_t       match_found = 0;
   int            r, e;

   /* sequential iteration: */
   for (r = hdb_seqinit(H); r != 0; r = hdb_seqnext(H)) {
      if (r == -1) {
         fatal_syserr("failure during hdb_seq()");
      }
      klen = hdb_klen(H);
      dlen = hdb_dlen(H);

      dynbuf_CLEAR(&K);
      if (hdb_dynread(H, &K, klen, hdb_kpos(H))) {
         fatal_syserr("failure during hdb_dynread()");
      }
      dynbuf_putnul(&K);
      key = (char *) dynbuf_BUF(&K);

      dynbuf_CLEAR(&D);
      if (hdb_dynread(H, &D, dlen, hdb_dpos(H))) {
         fatal_syserr("failure during hdb_dynread()");
      }
      dynbuf_putnul(&D);
      data = (char *) dynbuf_BUF(&D);

      e = regexec(&rx, (opt_k ? key : data), 0, NULL, 0);

      /* conditions skipping output for this record: */
      if (                      /* no match, not inverted: */
            ((e == REG_NOMATCH) && !opt_NOT) ||
            /* match, but inverted: */
            ((e == 0) && opt_NOT)) {
         continue;
      } else if (e) {
         /* XXX, error regexec() */
         warn("failure regexec() on ", opt_k ? "key: " : "data: ",
              opt_k ? key : data);
         continue;
      }

      /* else match: */
      ++match_found;

      /* start output: */
      if (!opt_g) {
         if (opt_y)
            ioq_puts(ioq1, "+");
         ioq_vputs(ioq1, nfmt_uint32(nfmt, klen), opt_y ? "," : ":");
         ioq_vputs(ioq1, nfmt_uint32(nfmt, dlen), opt_y ? ":" : "\t");
      }

      /* output key: */
      ioq_put(ioq1, (uchar_t *) key, klen);

      /* output delimiter: */
      if (!opt_g) {
         ioq_puts(ioq1, opt_y ? "->" : ":");
      } else {
         ioq_put(ioq1, (uchar_t *) arg_d, 1);
      }

      /* output data: */
      ioq_put(ioq1, (uchar_t *) data, dlen);

      /* record terminal: */
      ioq_puts(ioq1, "\n");
   }

   /* empty line terminal for cdb legacy output: */
   if (opt_y)
      ioq_puts(ioq1, "\n");

   ioq_flush(ioq1);

   dynbuf_freebuf(&K);
   dynbuf_freebuf(&D);

   return (match_found ? 0 : 1);
}


int HDBAPP_MAIN(int argc, char *argv[])
{
   nextopt_t      nopt;
   char           opt;
   const char    *regex = NULL;
   int            rxflags = 0;
   struct hdb     H;
   int            is_stdin = 0;
   int            e;

#ifndef HDBAPP_MODULE
/* standalone options/initialization: */
   usage_prog = "[-hV] [-d del | -g | -y] [-i] [-k] [-!] regex [hdb]";
   nextopt_init(&nopt, argc, argv, ":hVd:gyik!");
   progname = nextopt_progname(&nopt);
#else
/* module options/initialization: */
   usage_mode = "[-h] [-d del | -g | -y] [-i] [-k] [-!] regex [hdb]";
   nextopt_init(&nopt, argc, argv, ":hd:gyik!");
#endif

   while ((opt = nextopt(&nopt))) {
      char           optc[2] = { nopt.opt_got, '\0' };
      switch (opt) {
      case 'h':
         usage();
         die(0);
         break;
      case 'V':
         version();
         die(0);
         break;
      case '!':
         ++opt_NOT;
         break;
      case 'd':
         ++opt_g;
         arg_d = nopt.opt_arg;
         break;
      case 'g':
         ++opt_g;
         break;
      case 'i':
         ++opt_i;
         break;
      case 'k':
         ++opt_k;
         break;
      case 'y':
         ++opt_y;
         break;
      case ':':
         fatal_usage("missing argument for option -", optc);
         break;
      case '?':
         if (nopt.opt_got != '?') {
            fatal_usage("invalid option -", optc);
         }
         /* else fallthrough: */
      default:
         die_usage();
         break;
      }
   }

   argc -= nopt.arg_ndx;
   argv += nopt.arg_ndx;

   if ((argc > 0) && (argv[0] != NULL) && (*argv[0] != '\0')) {
      regex = argv[0];
      --argc;
      ++argv;
   } else {
      fatal_usage("missing or empty regex argument");
   }

   if (opt_g && opt_y) {
      fatal_usage("options -g and -y are mutually exclusive");
   }

   rxflags = REG_EXTENDED | REG_NOSUB;
   if (opt_i)
      rxflags |= REG_ICASE;

   /* compile regex: */
   e = regcomp(&rx, regex, rxflags);
   if (e != 0) {
      switch (e) {
      case REG_ESPACE:
         fatal(111, "failure compiling regex: out of memory");
         break;
      default:
         fatal_usage("bad regex pattern");
      }
   }

   if ((argc > 0) && (argv[0] != NULL)) {
      /* hdb file argument: */
      if (hdb_open(&H, argv[0]) == -1) {
         fatal_syserr("failure initializing hdb from ", argv[0]);
      }
   } else {
      /* read from stdin: */
      if (hdb_init(&H, 0) == -1) {
         fatal_syserr("failure initializing hdb on stdin");
      }
      ++is_stdin;
   }

   /* note: do_grep() just aborts on failure */
   e = do_grep(&H);

   /* cleanup: */
   regfree(&rx);
   if (is_stdin) {
      hdb_clear(&H);
   } else {
      hdb_close(&H);
   }

   return e;
}


/* eof (hdbapp_grep.c) */
