/* hdbapp_keys.c
** hdb keys: write out keys in hdb file
** wcm, 2010.05.30 - 2010.12.14
** ===
*/

/* libc: */
#include <stdlib.h>

/* unix: */
#include <unistd.h>
#include <errno.h>

/* lasagna: */
#include "buf.h"
#include "cstr.h"
#include "die.h"
#include "ioq.h"
#include "ioq_std.h"
#include "nextopt.h"
#include "nfmt.h"
#include "sysstr.h"

/* lasagna hdb reader: */
#include "hdb.h"

/* hdbapp: */
#include "hdbapp.h"


#ifndef HDBAPP_MODULE
/* configuration for standalone program: */
/* these globals setup in main(): */
const char    *progname = NULL;
const char    *usage_prog = NULL;
/* these globals remain NULL in standalone: */
const char    *progmode = NULL;
const char    *usage_mode = NULL;
/* entry point for standalone: */
#define  HDBAPP_MAIN  main
#else
/* entry point for module: */
#define  HDBAPP_MAIN  hdbapp_keys
#endif                          /* #ifndef HDBAPP_MODULE */


/* objects in local scope: */
static char    nfmt[NFMT_SIZE];

/* -y: cdb "legacy" output format */
static int     opt_y = 0;
/* -g: "getline" output format */
static int     opt_g = 0;
/* -X: display hash (in hex) */
static int     opt_X = 0;

/* functions in local scope: */
static int     do_keys(struct hdb *H);


/* do_keys()
**   iterate and output each key in hdb 
**   abort on failure
*/
static
int do_keys(struct hdb *H)
{
   uint32_t       klen;
   dynbuf         K = dynbuf_INIT();
   uchar_t       *key;
   int            r;

   /* sequential iteration: */
   for (r = hdb_seqinit(H); r != 0; r = hdb_seqnext(H)) {
      if (r == -1) {
         fatal_syserr("failure during hdb_seq()");
      }

      dynbuf_CLEAR(&K);

      klen = hdb_klen(H);
      if (hdb_dynread(H, &K, klen, hdb_kpos(H))) {
         fatal_syserr("failure during hdb_dynread()");
      }
      key = (uchar_t *) dynbuf_BUF(&K);

      /* output klen: */
      if (!opt_g) {
         if (opt_y)
            ioq_puts(ioq1, "+");
         ioq_vputs(ioq1, nfmt_uint32(nfmt, klen), ":");
      }

      /* output key: */
      ioq_put(ioq1, key, klen);

      /* output hash: */
      if (opt_X) {
         uint32_t       hash = hdb_hash(dynbuf_BUF(&K), klen);
         ioq_vputs(ioq1, "\t[0x", nfmt_uint32x_pad0(nfmt, hash, 8), "]");
      }
      ioq_puts(ioq1, "\n");
   }

   /* empty line terminal for cdb "legacy" output: */
   if (opt_y)
      ioq_puts(ioq1, "\n");

   ioq_flush(ioq1);

   dynbuf_freebuf(&K);
   return 0;
}


int HDBAPP_MAIN(int argc, char *argv[])
{
   nextopt_t      nopt;
   char           opt;
   struct hdb     H;
   int            is_stdin = 0;
   int            e;

#ifndef HDBAPP_MODULE
/* standalone options/initialization: */
   usage_prog = "[-hV] [-g | -y] [-X] [hdb]";
   nextopt_init(&nopt, argc, argv, ":hVgyX");
   progname = nextopt_progname(&nopt);
#else
/* module options/initialization: */
   usage_mode = "[-h] [-g | -y] [-X] [hdb]";
   nextopt_init(&nopt, argc, argv, ":hgyX");
#endif

   while ((opt = nextopt(&nopt))) {
      char           optc[2] = { nopt.opt_got, '\0' };
      switch (opt) {
      case 'h':
         usage();
         die(0);
         break;
      case 'V':
         version();
         die(0);
         break;
      case 'g':
         ++opt_g;
         break;
      case 'y':
         ++opt_y;
         break;
      case 'X':
         ++opt_X;
         break;
      case ':':
         fatal_usage("missing argument for option -", optc);
         break;
      case '?':
         if (nopt.opt_got != '?') {
            fatal_usage("invalid option -", optc);
         }
         /* else fallthrough: */
      default:
         die_usage();
         break;
      }
   }

   argc -= nopt.arg_ndx;
   argv += nopt.arg_ndx;

   if (opt_g && opt_y) {
      fatal_usage("options -g and -y are mutually exclusive");
   }

   if ((argc > 0) && (argv[0] != NULL)) {
      /* hdb file argument: */
      if (hdb_open(&H, argv[0]) == -1) {
         fatal_syserr("failure initializing hdb from ", argv[0]);
      }
   } else {
      /* read from stdin: */
      if (hdb_init(&H, 0) == -1) {
         fatal_syserr("failure initializing hdb on stdin");
      }
      ++is_stdin;
   }

   /* note: do_keys() just aborts on failure */
   e = do_keys(&H);

   /* close/cleanup: */
   if (is_stdin) {
      hdb_clear(&H);
   } else {
      hdb_close(&H);
   }

   return e;
}


/* eof (hdbapp_keys.c) */
