/* hdbapp_purge.c
** hdb purge module
** wcm, 2010.11.02 - 2010.12.14
** ===
*/

/* libc: */
#include <stdlib.h>
/* prototype for rename() in stdio.h: */
extern int     rename(const char *from, const char *to);

/* unix: */
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

/* lasagna: */
#include "buf.h"
#include "cstr.h"
#include "ioq.h"
#include "ioq_std.h"
#include "nextopt.h"
#include "nfmt.h"
#include "nuscan.h"
#include "sysstr.h"
#include "uchar.h"
#include "upak.h"

/* lasagna hdb reader: */
#include "hdb.h"
/* lasagna hdb creator: */
#include "hdbmk.h"

/* hdbapp: */
#include "hdbapp.h"


#ifndef HDBAPP_MODULE
/* configuration for standalone program: */
/* these globals setup in main(): */
const char    *progname = NULL;
const char    *usage_prog = NULL;
/* these globals remain NULL in standalone: */
const char    *progmode = NULL;
const char    *usage_mode = NULL;
/* entry point for standalone: */
#define  HDBAPP_MAIN  main
#else
/* entry point for module: */
#define  HDBAPP_MAIN  hdbapp_purge
#endif                          /* #ifndef HDBAPP_MODULE */


/*
** objects in local scope - options:
*/


/*
** objects in local scope - other:
*/
static const char *tmpext = HDBAPP_TMPEXT;
/* M = A minus B */
struct hdbmk   M;
struct hdb     hdb_A, hdb_B;

/*
** functions in local scope:
*/
static int     do_purge(void);


/* do_purge()
**   into initialized hdbmk object M, "subtract" B from A
**   ie, omit records from A with matching keys in B
**   abort on failure
*/
static
int do_purge()
{
   struct hdb    *A, *B;
   uint32_t       kpos, klen;
   dynbuf         K = dynbuf_INIT();
   int            r;

   A = &hdb_A;
   B = &hdb_B;

   /* scan A records: */
   for (r = hdb_seqinit(A); r != 0; r = hdb_seqnext(A)) {
      if (r == -1) {
         fatal_syserr("failure during hdb_seq()");
      }

      /* omit records from A with key found in B: */
      klen = hdb_klen(A);
      kpos = hdb_kpos(A);

      /* XXX, optimize for mmap()'d A */
      dynbuf_CLEAR(&K);
      if (hdb_dynread(A, &K, klen, kpos) == -1) {
         fatal_syserr("failure during hdb_dynread()");
      }

      /* lookup A key in B: */
      switch (hdb_find(B, dynbuf_BUF(&K), klen)) {
      case -1:
         fatal_syserr("failure during hdb_find()");
         break;
      case 1:
         /* match, skip this record in A: */
         continue;
      }

      /* output current record in A: */
      if (hdbmk_addrec(&M, A) == -1) {
         fatal_syserr("failure during hdbmk_addrec()");
      }
   }

   /* finish make: */
   if (hdbmk_finish(&M) == -1) {
      fatal_syserr("failure during hdbmk_finish()");
   }

   /* all done: */
   return 0;
}


int HDBAPP_MAIN(int argc, char *argv[])
{
   nextopt_t      nopt;
   char           opt;
   uint32_t       u;
   const char    *z;
   int            fd = -1, perms = -1;
   char          *outfile = NULL, *tmpfile = NULL;
   char          *fn_tmp = NULL;        /* non-NULL if allocated */
   char          *hdbA_file, *hdbB_file;
   char          *comment = NULL;
   uint32_t       clen = 0;
   int            e;

#ifndef HDBAPP_MODULE
/* standalone options/initialization: */
   usage_prog =
       "[-hV] [-c comment] [-p perms] [-t tmpfile] result.hdb A.hdb B.hdb";
   nextopt_init(&nopt, argc, argv, ":hVc:p:t:");
   progname = nextopt_progname(&nopt);
#else
/* module options/initialization: */
   usage_mode =
       "[-h] [-c comment] [-p perms] [-t tmpfile] result.hdb A.hdb B.hdb";
   nextopt_init(&nopt, argc, argv, ":hc:p:t:");
#endif

   while ((opt = nextopt(&nopt))) {
      char           optc[2] = { nopt.opt_got, '\0' };
      switch (opt) {
      case 'h':
         usage();
         die(0);
         break;
      case 'V':
         version();
         die(0);
         break;
      case 'c':
         comment = nopt.opt_arg;
         break;
      case 'p':
         z = nuscan_uint32o(&u, nopt.opt_arg);
         if (*z != '\0') {
            fatal_usage("non-numeric/octal argument found for option -", optc,
                        ": ", nopt.opt_arg);
         }
         perms = (int) u;
         break;
      case ':':
         fatal_usage("missing argument for option -", optc);
         break;
      case '?':
         if (nopt.opt_got != '?') {
            fatal_usage("invalid option -", optc);
         }
         /* else fallthrough: */
      default:
         die_usage();
         break;
      }
   }

   argc -= nopt.arg_ndx;
   argv += nopt.arg_ndx;

   if ((argc > 0) && (argv[0] != NULL) && (*argv[0] != '\0')) {
      /* result_file argument: */
      outfile = argv[0];
      ++argv;
      --argc;
   } else {
      fatal_usage("missing \"result.hdb\" file argument");
   }

   if ((argc > 0) && (argv[0] != NULL) && (*argv[0] != '\0')) {
      /* hdbA_file argument: */
      hdbA_file = argv[0];
      ++argv;
      --argc;
   } else {
      fatal_usage("missing \"A.hdb\" file argument");
   }

   if ((argc > 0) && (argv[0] != NULL) && (*argv[0] != '\0')) {
      /* hdbB_file argument: */
      hdbB_file = argv[0];
      ++argv;
      --argc;
   } else {
      fatal_usage("missing \"B.hdb\" file argument");
   }

   if (tmpfile == NULL) {
      fn_tmp = (char *) malloc(cstr_len(outfile) + cstr_len(tmpext) + 1);
      if (fn_tmp == NULL) {
         fatal_syserr("failure malloc() for naming tmpfile");
      }
      cstr_vcopy(fn_tmp, outfile, tmpext);
      tmpfile = fn_tmp;
   }

   if (perms >= 0)
      umask(0);
   else
      perms = 0666;

   unlink(tmpfile);
   fd = open(tmpfile, O_WRONLY | O_CREAT | O_EXCL, (mode_t) perms);
   if (fd == -1) {
      fatal_syserr("failure on open() for ", tmpfile);
   }

   clen = (uint32_t) (comment == NULL ? 0 : cstr_len(comment));
   if (hdbmk_start(&M, fd, (const uchar_t *) comment, clen) == -1) {
      fatal_syserr("failure hdbmk_start()");
   }

   if (hdb_open(&hdb_A, hdbA_file) == -1) {
      fatal_syserr("failure hdb_open() on ", hdbA_file);
   }

   if (hdb_open(&hdb_B, hdbB_file) == -1) {
      fatal_syserr("failure hdb_open() on ", hdbB_file);
   }

   e = do_purge();

   if (e == 0) {
      if (fsync(fd) == -1) {
         fatal_syserr("failure fsync()");
      }

      if (close(fd) == -1) {
         fatal_syserr("failure close()");
      }

      if (rename(tmpfile, outfile) == -1) {
         fatal_syserr("failure rename()");
      }
   }

   /* gratuitous cleanup: */
   hdb_close(&hdb_A);
   hdb_close(&hdb_B);
   hdbmk_clear(&M);
   if (fn_tmp != NULL)
      free(fn_tmp);

   return e;
}


/* eof (hdbapp_purge.c) */
